"""
Compressors are basically serializers from `bytes` to `bytes`
that are grouped for convenience.
"""

from .zlib import ZlibCompressor  # noqa

try:
    from .zstd import ZstdCompressor  # noqa
except ImportError:
    pass  # only available with `zstd` feature



from typing import Any, Callable


def make_default_key(callable_: Callable[..., Any], *args: Any, **kwargs: Any) -> str:
    """
    Generates cache key given the callable and the arguments it's being called with.
    """

    # noinspection PyUnresolvedReferences
    parts = (
        callable_.__module__,
        callable_.__qualname__,
        # Since we join with `:`, we need to «escape» `:`s with `::`.
        *(str(arg).replace(":", "::") for arg in args),
        *(f"{str(key).replace(':', '::')}={str(value).replace(':', '::')}" for key, value in sorted(kwargs.items())),
    )
    return ":".join(parts)



"""
Sync cache backend protocol descriptors.

Note, that some methods provide reasonable default implementation,
but implementors SHOULD override them for the sake of performance.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from contextlib import AbstractContextManager
from datetime import datetime, timedelta
from typing import Generic, Iterable, Optional, Tuple

from typing_extensions import Protocol

from cachetory.interfaces.backends.private import WireT, WireT_co, WireT_contra
from cachetory.private.datetime import make_deadline


class SyncBackendRead(Protocol[WireT_co]):
    """
    Describes the read operations of a synchronous cache.
    """

    def get(self, key: str) -> WireT_co:  # pragma: no cover
        """
        Retrieve a value from the cache.

        Returns:
            Cached value, if it exists.
        Raises:
            KeyError: the key does not exist.
        """
        raise NotImplementedError

    def get_many(self, *keys: str) -> Iterable[Tuple[str, WireT_co]]:
        """
        Get all the values corresponding to the specified keys.

        Returns:
            Existing key-value pairs.
        """
        for key in keys:
            try:
                value = self.get(key)
            except KeyError:
                pass
            else:
                yield key, value


class SyncBackendWrite(Protocol[WireT_contra]):
    """
    Describes the write operations of a synchronous cache.
    """

    def expire_in(self, key: str, time_to_live: Optional[timedelta] = None) -> None:
        """
        Set the expiration time on the key.
        """
        self.expire_at(key, make_deadline(time_to_live))

    def expire_at(self, key: str, deadline: Optional[datetime]) -> None:  # pragma: no cover
        """
        Set the expiration deadline on the key.
        """
        raise NotImplementedError

    def set(
        self,
        key: str,
        value: WireT_contra,
        *,
        time_to_live: Optional[timedelta] = None,
        if_not_exists: bool = False,
    ) -> bool:  # pragma: no cover
        """
        Put the value into the cache.

        Returns:
            `True` if the value has been successfully set, `False` when `if_not_exists` is true
            and the key is already existing.
        """
        raise NotImplementedError

    def set_many(self, items: Iterable[Tuple[str, WireT_contra]]) -> None:
        """
        Put all the specified values to the cache.
        """
        for (key, value) in items:
            self.set(key, value)

    def delete(self, key: str) -> bool:  # pragma: no cover
        """
        Delete the key from the cache.

        Returns:
            `True` if the key has existed, `False` otherwise.
        """
        raise NotImplementedError

    def clear(self) -> None:  # pragma: no cover
        """
        Clears the backend storage.
        """
        raise NotImplementedError


class SyncBackend(
    AbstractContextManager,
    SyncBackendRead[WireT],
    SyncBackendWrite[WireT],
    Generic[WireT],
    metaclass=ABCMeta,
):
    """
    Generic synchronous cache backend.
    """

    @classmethod
    @abstractmethod
    def from_url(cls, url: str) -> SyncBackend:  # pragma: no cover
        """
        Create a synchronous cache backend from the specified URL.

        Returns:
            An instance of the specific backend class.
        """
        raise NotImplementedError

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        return None



from cachetory.serializers.compressors.zstd import ZstdCompressor


def test_serialize_deserialize():
    serializer = ZstdCompressor()
    value = b"hello, world!"
    assert serializer.deserialize(serializer.serialize(value)) == value



from __future__ import annotations

import itertools
from datetime import datetime, timedelta
from typing import Iterable, Optional, Tuple

import redis

from cachetory.interfaces.backends.sync import SyncBackend


class RedisBackend(SyncBackend[bytes]):
    """
    Synchronous Redis (`redis-py`) backend.
    """

    @classmethod
    def from_url(cls, url: str) -> RedisBackend:
        if url.startswith("redis+"):
            url = url[6:]
        return cls(redis.Redis.from_url(url))

    def __init__(self, client: redis.Redis):
        self._client = client

    def get(self, key: str) -> bytes:
        data = self._client.get(key)
        if data is not None:
            return data
        raise KeyError(key)

    def get_many(self, *keys: str) -> Iterable[Tuple[str, bytes]]:
        for key, value in zip(keys, self._client.mget(*keys)):
            if value is not None:
                yield key, value

    def expire_at(self, key: str, deadline: Optional[datetime]) -> None:
        if deadline:
            # One can pass `datetime` directly to `pexpireat`, but the latter
            # incorrectly converts datetime into timestamp.
            self._client.pexpireat(key, int(deadline.timestamp() * 1000.0))
        else:
            self._client.persist(key)

    def expire_in(self, key: str, time_to_live: Optional[timedelta] = None) -> None:
        if time_to_live:
            self._client.pexpire(key, time_to_live)
        else:
            self._client.persist(key)

    def set(
        self,
        key: str,
        value: bytes,
        *,
        time_to_live: Optional[timedelta] = None,
        if_not_exists: bool = False,
    ) -> bool:
        return bool(self._client.set(key, value, px=time_to_live, nx=if_not_exists))

    def set_many(self, items: Iterable[Tuple[str, bytes]]) -> None:
        self._client.execute_command("MSET", *itertools.chain.from_iterable(items))

    def delete(self, key: str) -> bool:
        return bool(self._client.delete(key))

    def clear(self) -> None:
        self._client.flushdb()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        return self._client.__exit__(exc_type, exc_value, traceback)



from pytest import fixture

from cachetory import serializers
from cachetory.backends import sync as sync_backends
from cachetory.caches.sync import Cache
from tests.support import if_redis_enabled


@fixture
def memory_cache() -> Cache[int]:
    return Cache(
        serializer=serializers.from_url("pickle+zstd://?pickle-protocol=4&compression-level=3"),
        backend=sync_backends.from_url("memory://"),
    )


def test_get_set_in_memory(memory_cache: Cache[int]):
    memory_cache.set("foo", 42)
    assert memory_cache.get("foo") == 42


def test_set_item(memory_cache: Cache[int]):
    memory_cache["foo"] = 42
    assert memory_cache.get("foo") == 42


def test_get_default(memory_cache: Cache[int]):
    assert memory_cache.get("missing", 100500) == 100500


def test_get_many(memory_cache: Cache[int]):
    memory_cache.set("foo", 42)
    assert memory_cache.get_many("foo", "bar") == {"foo": 42}


def test_set_many(memory_cache: Cache[int]):
    memory_cache.set_many({"foo": 42, "bar": 100500})
    assert memory_cache.get("foo") == 42
    assert memory_cache.get("bar") == 100500


def test_delete(memory_cache: Cache[int]):
    memory_cache.set("foo", 42)
    assert memory_cache.delete("foo")
    assert not memory_cache.delete("foo")
    assert memory_cache.get("foo") is None


def test_del_item(memory_cache: Cache[int]):
    memory_cache.set("foo", 42)
    del memory_cache["foo"]
    assert memory_cache.get("foo") is None


@if_redis_enabled
def test_get_set_in_redis():
    cache = Cache[int](
        serializer=serializers.from_url("pickle+zlib://"),
        backend=sync_backends.from_url("redis://localhost:6379"),
    )
    with cache:
        cache.set("foo", 42)
        assert cache.get("foo") == 42



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any, Coroutine, Generic, Optional

from cachetory.backends.sync.memory import MemoryBackend as SyncMemoryBackend
from cachetory.interfaces.backends.async_ import AsyncBackend
from cachetory.interfaces.backends.private import WireT
from cachetory.private.asyncio import postpone


class MemoryBackend(AsyncBackend[WireT], Generic[WireT]):
    """
    Memory backend that stores everything in a local dictionary.
    """

    __slots__ = ("_inner",)

    @classmethod
    def from_url(cls, _url: str) -> MemoryBackend:
        return MemoryBackend()

    def __init__(self):
        # We'll simply delegate call to the wrapped backend.
        self._inner = SyncMemoryBackend()

    def get(self, key: str) -> Coroutine[Any, Any, WireT]:
        return postpone(self._inner.get, key)

    def expire_at(self, key: str, deadline: Optional[datetime]) -> Coroutine[Any, Any, None]:
        return postpone(self._inner.expire_at, key, deadline)

    def set(
        self,
        key: str,
        value: WireT,
        *,
        time_to_live: Optional[timedelta] = None,
        if_not_exists: bool = False,
    ) -> Coroutine[Any, Any, bool]:
        return postpone(
            self._inner.set,
            key,
            value,
            time_to_live=time_to_live,
            if_not_exists=if_not_exists,
        )

    def delete(self, key: str) -> Coroutine[Any, Any, bool]:
        return postpone(self._inner.delete, key)

    def clear(self) -> Coroutine[Any, Any, None]:
        return postpone(self._inner.clear)

    @property
    def size(self) -> int:
        return self._inner.size



from typing import Callable, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")


async def postpone(f: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
    """
    Postpones `f` until awaited and forwards the value back to the caller.
    """
    return f(*args, **kwargs)



from cachetory.interfaces.backends.private import WireT
from cachetory.interfaces.serializers import Serializer, ValueT

from .chained import ChainedSerializer
from .noop import NoopSerializer  # noqa
from .pickle import PickleSerializer  # noqa


def from_url(url: str) -> Serializer[ValueT, WireT]:
    return ChainedSerializer[ValueT, WireT].from_url(url)



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Generic, Iterable, Optional, Tuple

from cachetory.interfaces.backends.private import WireT
from cachetory.interfaces.backends.sync import SyncBackend


class DummyBackend(SyncBackend[WireT], Generic[WireT]):
    """
    Dummy backend that stores nothing.
    """

    @classmethod
    def from_url(cls, url: str) -> DummyBackend:
        return DummyBackend()

    def get(self, key: str) -> WireT:  # pragma: no cover
        raise KeyError(key)

    def get_many(self, *keys: str) -> Iterable[Tuple[str, WireT]]:  # pragma: no cover
        return []

    def expire_in(self, key: str, time_to_live: Optional[timedelta] = None) -> None:  # pragma: no cover
        return None

    def expire_at(self, key: str, deadline: Optional[datetime]) -> None:  # pragma: no cover
        return None

    def set(
        self,
        key: str,
        value: WireT,
        *,
        time_to_live: Optional[timedelta] = None,
        if_not_exists: bool = False,
    ) -> bool:  # pragma: no cover
        return True

    def set_many(self, items: Iterable[Tuple[str, WireT]]) -> None:  # pragma: no cover
        return None

    def delete(self, key: str) -> bool:  # pragma: no cover
        return False  # has never been there

    def clear(self) -> None:  # pragma: no cover
        return None  # already perfectly clean



from contextlib import AbstractContextManager
from datetime import timedelta
from typing import Dict, Generic, Iterable, Mapping, Optional, Tuple, Union

from cachetory.caches.private import DefaultT
from cachetory.interfaces.backends.private import WireT
from cachetory.interfaces.backends.sync import SyncBackend
from cachetory.interfaces.serializers import Serializer, ValueT


class Cache(AbstractContextManager, Generic[ValueT]):
    """
    Synchronous cache.
    """

    __slots__ = ("_serializer", "_backend")

    def __init__(self, *, serializer: Serializer[ValueT, WireT], backend: SyncBackend[WireT]):
        self._serializer = serializer
        self._backend = backend

    def __getitem__(self, key: str) -> ValueT:
        """
        Retrieve a single value from the cache.

        Raises:
            KeyError: the key is not found in the cache
        """
        return self._serializer.deserialize(self._backend.get(key))

    def get(self, key: str, default: DefaultT = None) -> Union[ValueT, DefaultT]:  # type: ignore
        """
        Retrieve a single value from the cache.

        Args:
            key: cache key
            default: default value – if the key is not found

        Returns:
            Retrieved value if present, or `default` otherwise.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def get_many(self, *keys: str) -> Dict[str, ValueT]:
        """
        Retrieve many values from the cache.

        Returns:
            Dictionary of existing values indexed by their respective keys.
            Missing keys are omitted.
        """
        return {key: self._serializer.deserialize(data) for key, data in self._backend.get_many(*keys)}

    def expire_in(self, key: str, time_to_live: Optional[timedelta] = None) -> None:
        return self._backend.expire_in(key, time_to_live)

    def __setitem__(self, key: str, value: ValueT) -> None:
        self._backend.set(key, self._serializer.serialize(value), time_to_live=None)

    def set(
        self,
        key: str,
        value: ValueT,
        time_to_live: Optional[timedelta] = None,
        if_not_exists: bool = False,
    ) -> None:
        self._backend.set(
            key,
            self._serializer.serialize(value),
            time_to_live=time_to_live,
            if_not_exists=if_not_exists,
        )

    def set_many(self, items: Union[Iterable[Tuple[str, ValueT]], Mapping[str, ValueT]]) -> None:
        if isinstance(items, Mapping):
            items = items.items()
        self._backend.set_many((key, self._serializer.serialize(value)) for key, value in items)

    def delete(self, key: str) -> bool:
        return self._backend.delete(key)

    def __delitem__(self, key: str) -> None:
        self._backend.delete(key)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        return self._backend.__exit__(exc_type, exc_val, exc_tb)



from asyncio import get_running_loop
from concurrent.futures import Executor
from contextlib import AbstractAsyncContextManager
from datetime import timedelta
from typing import Dict, Generic, Iterable, Mapping, Optional, Tuple, Union

from cachetory.caches.private import DefaultT
from cachetory.interfaces.backends.async_ import AsyncBackend
from cachetory.interfaces.backends.private import WireT
from cachetory.interfaces.serializers import Serializer, ValueT
from cachetory.private.typing import NotSet

_NOT_SET = NotSet()


class Cache(AbstractAsyncContextManager, Generic[ValueT]):
    """
    Asynchronous cache.
    """

    __slots__ = ("_serializer", "_backend", "_serialize_executor")

    def __init__(
        self,
        *,
        serializer: Serializer[ValueT, WireT],
        backend: AsyncBackend[WireT],
        serialize_executor: Union[None, Executor, NotSet] = _NOT_SET,
    ):
        """
        Args:
            serialize_executor:
                If specified, underlying serializing and deserializing will be performed
                using the executor (for example, ``concurrent.futures.ProcessPoolExecutor``).
                This may be useful to better utilize CPU when caching large blobs.
                If not specified, (de)serialization is performed in the current thread.
        """
        self._serializer = serializer
        self._backend = backend
        self._serialize_executor = serialize_executor

    async def get(self, key: str, default: DefaultT = None) -> Union[ValueT, DefaultT]:  # type: ignore
        try:
            data = await self._backend.get(key)
        except KeyError:
            return default
        else:
            return await self._deserialize(data)

    async def get_many(self, *keys: str) -> Dict[str, ValueT]:
        return {key: await self._deserialize(data) async for key, data in self._backend.get_many(*keys)}

    async def expire_in(self, key: str, time_to_live: Optional[timedelta] = None) -> None:
        return await self._backend.expire_in(key, time_to_live)

    async def set(
        self,
        key: str,
        value: ValueT,
        time_to_live: Optional[timedelta] = None,
        if_not_exists: bool = False,
    ) -> None:
        await self._backend.set(
            key,
            await self._serialize(value),
            time_to_live=time_to_live,
            if_not_exists=if_not_exists,
        )

    async def set_many(self, items: Union[Iterable[Tuple[str, ValueT]], Mapping[str, ValueT]]) -> None:
        if isinstance(items, Mapping):
            items = items.items()
        await self._backend.set_many([(key, await self._serialize(value)) for key, value in items])

    async def delete(self, key: str) -> bool:
        return await self._backend.delete(key)

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        return await self._backend.__aexit__(exc_type, exc_val, exc_tb)

    async def _serialize(self, value: ValueT) -> WireT:
        if self._serialize_executor is _NOT_SET:
            return self._serializer.serialize(value)
        else:
            return await get_running_loop().run_in_executor(self._serialize_executor, self._serializer.serialize, value)

    async def _deserialize(self, data: WireT) -> ValueT:
        if self._serialize_executor is _NOT_SET:
            return self._serializer.deserialize(data)
        else:
            return await get_running_loop().run_in_executor(
                self._serialize_executor, self._serializer.deserialize, data
            )



from __future__ import annotations

import zlib
from urllib.parse import parse_qsl, urlparse

from pydantic import BaseModel, Field, conint

from cachetory.interfaces.serializers import Serializer


class ZlibCompressor(Serializer[bytes, bytes]):
    """
    Uses the built-in zlib compression.
    """

    __slots__ = ("_level",)

    @classmethod
    def from_url(cls, url: str) -> ZlibCompressor:
        params = _UrlParams.parse_obj(dict(parse_qsl(urlparse(url).query)))
        return cls(compression_level=params.compression_level)

    def __init__(self, *, compression_level: int = zlib.Z_DEFAULT_COMPRESSION):
        self._level = compression_level

    def serialize(self, value: bytes) -> bytes:
        return zlib.compress(value, level=self._level)

    def deserialize(self, data: bytes) -> bytes:
        return zlib.decompress(data)


class _UrlParams(BaseModel):
    compression_level: conint(ge=-1, le=9) = Field(  # type: ignore
        zlib.Z_DEFAULT_COMPRESSION,
        alias="compression-level",
    )



from datetime import datetime, timedelta, timezone
from typing import Iterable

from freezegun import freeze_time
from pytest import fixture, raises

from cachetory.backends.sync.memory import MemoryBackend


@fixture
def backend() -> Iterable[MemoryBackend[int]]:
    with MemoryBackend[int]() as backend:
        yield backend


def test_get_existing(backend: MemoryBackend[int]):
    backend.set("foo", 42)
    assert backend.get("foo") == 42


def test_get_missing(backend: MemoryBackend[int]):
    with raises(KeyError):
        backend.get("foo")


def test_set_default(backend: MemoryBackend[int]):
    assert backend.set("foo", 42, if_not_exists=True)
    assert not backend.set("foo", 43, if_not_exists=True)
    assert backend.get("foo") == 42
    assert backend.size == 1


def test_delete_existing(backend: MemoryBackend[int]):
    backend.set("foo", 42)
    assert backend.delete("foo")
    with raises(KeyError):
        backend.get("foo")


def test_delete_missing(backend: MemoryBackend[int]):
    assert not backend.delete("foo")


def test_set_get_many(backend: MemoryBackend[int]):
    backend.set_many([("foo", 42), ("bar", 100500)])
    assert backend.size == 2
    assert list(backend.get_many("foo", "bar")) == [("foo", 42), ("bar", 100500)]


def test_set_with_ttl(backend: MemoryBackend[int]):
    with freeze_time("2022-06-11 21:33:00"):
        backend.set("foo", 42, time_to_live=timedelta(seconds=59))
    with freeze_time("2022-06-11 21:33:58"):
        assert backend.get("foo") == 42
    with freeze_time("2022-06-11 21:34:00"), raises(KeyError):
        assert backend.get("foo")
    assert backend.size == 0


def test_expire_at(backend: MemoryBackend[int]):
    backend.set("foo", 42)
    backend.expire_at("foo", datetime(2022, 6, 10, 21, 50, 00, tzinfo=timezone.utc))

    with freeze_time("2022-06-10 21:49:59"):
        assert backend.get("foo") == 42
    with freeze_time("2022-06-10 21:50:00"), raises(KeyError):
        assert backend.get("foo")
    assert backend.size == 0


def test_expire_in(backend: MemoryBackend[int]):
    with freeze_time("2022-06-10 21:49:00"):
        backend.set("foo", 42)
        backend.expire_in("foo", timedelta(seconds=59))
    with freeze_time("2022-06-10 21:50:00"), raises(KeyError):
        assert backend.get("foo")
    assert backend.size == 0


def test_clear(backend: MemoryBackend[int]):
    backend.set("foo", 42)
    backend.clear()
    assert backend.size == 0



from typing import Any

from pytest import fixture, mark

from cachetory.backends.async_ import MemoryBackend
from cachetory.caches.async_ import Cache
from cachetory.decorators.async_ import cached
from cachetory.serializers import NoopSerializer


@fixture
def cache() -> Cache[int]:
    return Cache(serializer=NoopSerializer(), backend=MemoryBackend[int]())


@fixture
def cache_2() -> Cache[int]:
    return Cache(serializer=NoopSerializer(), backend=MemoryBackend[int]())


@mark.asyncio
async def test_simple(cache: Cache[int]):
    call_counter = 0

    @cached(cache)
    async def expensive_function(arg: int, *, kwarg: int) -> int:
        assert arg == 1, "the positional argument is not forwarded"
        assert kwarg == 2, "the keyword argument is not forwarded"
        nonlocal call_counter
        call_counter += 1
        return 42

    assert await expensive_function(1, kwarg=2) == 42, "the return value is not forwarded"
    assert call_counter == 1
    await expensive_function(1, kwarg=2)
    assert call_counter == 1, "cache did not work"


@mark.asyncio
async def test_callable_cache(cache: Cache[int], cache_2: Cache[int]):
    call_counter = 0

    async def choose_cache(_wrapped_callable: Any, param: int) -> Cache[int]:
        return cache_2 if param == 2 else cache

    @cached(cache=choose_cache)
    async def expensive_function(_: int) -> int:
        nonlocal call_counter
        call_counter += 1
        return 42

    await expensive_function(1)
    assert call_counter == 1

    await expensive_function(2)
    assert call_counter == 2

    assert cache._backend.size == 1  # type: ignore
    assert cache_2._backend.size == 1  # type: ignore



from datetime import timedelta
from time import sleep
from typing import Iterable

from pytest import fixture, raises

from cachetory.backends.sync import RedisBackend
from cachetory.private.datetime import make_deadline
from tests.support import if_redis_enabled


@fixture
def backend() -> Iterable[RedisBackend]:
    with RedisBackend.from_url("redis://localhost:6379") as backend:
        backend.clear()
        try:
            yield backend
        finally:
            backend.clear()


@if_redis_enabled
def test_get_existing(backend: RedisBackend):
    backend.set("foo", b"hello")
    assert backend.get("foo") == b"hello"


@if_redis_enabled
async def test_get_missing(backend: RedisBackend):
    with raises(KeyError):
        backend.get("foo")


@if_redis_enabled
async def test_set_default(backend: RedisBackend):
    assert backend.set("foo", b"hello", if_not_exists=True)
    assert not backend.set("foo", b"world", if_not_exists=True)
    assert backend.get("foo") == b"hello"


@if_redis_enabled
async def test_delete_existing(backend: RedisBackend):
    backend.set("foo", b"hello")
    assert backend.delete("foo")
    with raises(KeyError):
        backend.get("foo")


@if_redis_enabled
async def test_delete_missing(backend: RedisBackend):
    assert not backend.delete("foo")


@if_redis_enabled
async def test_set_get_many(backend: RedisBackend):
    backend.set_many([("non-empty", b"foo"), ("empty", b"")])
    assert list(backend.get_many("non-empty", "missing", "empty")) == [("non-empty", b"foo"), ("empty", b"")]


@if_redis_enabled
async def test_set_with_ttl(backend: RedisBackend):
    backend.set("foo", b"bar", time_to_live=timedelta(seconds=0.25))
    assert backend.get("foo") == b"bar"
    sleep(0.5)
    with raises(KeyError):
        backend.get("foo")


@if_redis_enabled
async def test_expire_at(backend: RedisBackend):
    backend.set("foo", b"bar")
    backend.expire_at("foo", make_deadline(timedelta(seconds=0.25)))
    assert backend.get("foo") == b"bar"
    sleep(0.5)
    with raises(KeyError):
        backend.get("foo")


@if_redis_enabled
async def test_expire_in(backend: RedisBackend):
    backend.set("foo", b"bar")
    backend.expire_in("foo", timedelta(seconds=0.25))
    assert backend.get("foo") == b"bar"
    sleep(0.5)
    with raises(KeyError):
        backend.get("foo")


@if_redis_enabled
async def test_clear(backend: RedisBackend):
    backend.set("foo", b"bar")
    backend.clear()
    with raises(KeyError):
        backend.get("foo")


@if_redis_enabled
def test_get_empty_value(backend: RedisBackend):
    backend.set("foo", b"")
    assert backend.get("foo") == b""



from pytest import fixture, mark

from cachetory import serializers
from cachetory.backends import async_ as async_backends
from cachetory.caches.async_ import Cache
from tests.support import if_redis_enabled


@fixture
async def memory_cache() -> Cache[int]:
    return Cache(
        serializer=serializers.from_url("pickle+zstd://"),
        backend=async_backends.from_url("memory://"),
    )


@mark.asyncio
async def test_get_set_in_memory(memory_cache: Cache[int]):
    await memory_cache.set("foo", 42)
    assert await memory_cache.get("foo") == 42


@mark.asyncio
async def test_get_default(memory_cache: Cache[int]):
    assert await memory_cache.get("missing", 100500) == 100500


@mark.asyncio
async def test_get_many(memory_cache: Cache[int]):
    await memory_cache.set("foo", 42)
    assert await memory_cache.get_many("foo", "bar") == {"foo": 42}


@mark.asyncio
async def test_set_many(memory_cache: Cache[int]):
    await memory_cache.set_many({"foo": 42, "bar": 100500})
    assert await memory_cache.get("foo") == 42
    assert await memory_cache.get("bar") == 100500


@mark.asyncio
async def test_delete(memory_cache: Cache[int]):
    await memory_cache.set("foo", 42)
    assert await memory_cache.delete("foo")
    assert not await memory_cache.delete("foo")
    assert await memory_cache.get("foo") is None


@mark.asyncio
async def test_serialize_executor():
    cache = Cache(
        serializer=serializers.from_url("pickle://"),
        backend=async_backends.from_url("memory://"),
        serialize_executor=None,  # that's the default executor, NOT «no executor»
    )
    await cache.set("foo", 42)
    assert await cache.get("foo") == 42


@if_redis_enabled
@mark.asyncio
async def test_get_set_in_redis():
    cache = Cache[int](
        serializer=serializers.from_url("pickle+zlib://"),
        backend=async_backends.from_url("redis://localhost:6379"),
    )
    async with cache:
        await cache.set("foo", 42)
        assert await cache.get("foo") == 42



from __future__ import annotations

from urllib.parse import parse_qsl, urlparse

import zstd  # type: ignore
from pydantic import BaseModel, Field, conint

from cachetory.interfaces.serializers import Serializer


class ZstdCompressor(Serializer[bytes, bytes]):
    """
    Uses external `zstd` package for Zstandard compression.
    """

    __slots__ = ("_level", "_threads")

    @classmethod
    def from_url(cls, url: str) -> ZstdCompressor:
        params = _UrlParams.parse_obj(dict(parse_qsl(urlparse(url).query)))
        return cls(compression_level=params.compression_level, compression_threads=params.compression_threads)

    def __init__(
        self,
        *,
        compression_level: int = 3,
        compression_threads: int = 0,
    ):
        self._level = compression_level
        self._threads = compression_threads

    def serialize(self, value: bytes) -> bytes:
        return zstd.compress(value, self._level, self._threads)

    def deserialize(self, data: bytes) -> bytes:
        return zstd.decompress(data)


class _UrlParams(BaseModel):
    compression_level: int = Field(3, alias="compression-level")  # type: ignore
    compression_threads: conint(ge=0) = Field(0, alias="compression-threads")  # type: ignore



from datetime import timedelta
from functools import wraps
from typing import Callable, Optional, Union

from typing_extensions import ParamSpec

from cachetory.caches.sync import Cache
from cachetory.decorators import shared
from cachetory.interfaces.serializers import ValueT

P = ParamSpec("P")
"""
Original wrapped function parameter specification.
"""


def cached(
    cache: Union[Cache[ValueT], Callable[..., Cache[ValueT]]],  # no way to use `P` here
    *,
    make_key: Callable[..., str] = shared.make_default_key,  # no way to use `P` here
    time_to_live: Optional[timedelta] = None,
    if_not_exists: bool = False,
) -> Callable[[Callable[P, ValueT]], Callable[P, ValueT]]:
    """
    Args:
        cache:
            `Cache` instance or a callable tha returns a `Cache` instance for each function call.
            In the latter case the specified callable gets called with a wrapped function as the first argument,
            and the rest of the arguments next to it.
        make_key: callable to generate a custom cache key per each call.
        if_not_exists: controls concurrent sets: if `True` – avoids overwriting a cached value.
        time_to_live: cached value expiration time.
    """

    def wrap(callable_: Callable[P, ValueT]) -> Callable[P, ValueT]:
        @wraps(callable_)
        def cached_callable(*args: P.args, **kwargs: P.kwargs) -> ValueT:
            cache_ = cache(callable_, *args, **kwargs) if callable(cache) else cache
            key_ = make_key(callable_, *args, **kwargs)
            try:
                value = cache_[key_]
            except KeyError:
                value = callable_(*args, **kwargs)
                cache_.set(key_, value, time_to_live=time_to_live, if_not_exists=if_not_exists)
            return value

        return cached_callable

    return wrap



from typing import Any, Callable, Dict, Tuple

from pytest import mark

from cachetory.decorators.shared import make_default_key


def _callable():
    """
    This is only used as a test target.
    """


@mark.parametrize(
    "callable_, args, kwargs, expected_key",
    [
        (_callable, (), {}, "tests.decorators.test_shared:_callable"),
        (_callable, (1, 42), {}, "tests.decorators.test_shared:_callable:1:42"),
        (_callable, (), {"foo": 42, "bar": "qux"}, "tests.decorators.test_shared:_callable:bar=qux:foo=42"),
        (
            _callable,
            (),
            {"bar": "qux", "foo": 42},
            "tests.decorators.test_shared:_callable:bar=qux:foo=42",
        ),
        (_callable, (1, 42), {"foo": "bar"}, "tests.decorators.test_shared:_callable:1:42:foo=bar"),
        (
            _callable,
            ("a:b",),
            {"foo:bar": "qux:quux"},
            "tests.decorators.test_shared:_callable:a::b:foo::bar=qux::quux",
        ),
    ],
)
def test_make_default_key(
    callable_: Callable[..., Any],
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    expected_key: str,
):
    assert make_default_key(callable_, *args, **kwargs) == expected_key



from typing import List, Type

from pytest import mark, raises

from cachetory.interfaces.serializers import Serializer
from cachetory.serializers import ChainedSerializer, PickleSerializer
from cachetory.serializers.compressors import ZstdCompressor


@mark.parametrize(
    "url, expected_layers",
    [
        ("pickle://", [PickleSerializer]),
        ("pickle+zstd://", [PickleSerializer, ZstdCompressor]),
    ],
)
def test_layers(url: str, expected_layers: List[Type[Serializer]]):
    assert [type(layer) for layer in ChainedSerializer.from_url(url)._layers] == expected_layers


def test_unsupported_scheme():
    with raises(ValueError):
        ChainedSerializer.from_url("subspace://")


def test_serialize():
    serializer: Serializer[str, bytes] = ChainedSerializer.from_url("pickle+zstd://")
    value = "Shields up! Red alert!"
    assert serializer.serialize(value) == ZstdCompressor().serialize(PickleSerializer().serialize(value))


def test_deserialize():
    value = "Energize!"
    serialized_value = ZstdCompressor().serialize(PickleSerializer().serialize(value))
    assert ChainedSerializer.from_url("pickle+zstd://").deserialize(serialized_value) == value



from datetime import datetime, timedelta, timezone
from typing import AsyncIterable

from freezegun import freeze_time
from pytest import fixture, mark, raises

from cachetory.backends.async_.memory import MemoryBackend


@fixture
async def backend() -> AsyncIterable[MemoryBackend[int]]:
    async with MemoryBackend[int]() as backend:
        yield backend


@mark.asyncio
async def test_get_existing(backend: MemoryBackend[int]):
    await backend.set("foo", 42)
    assert await backend.get("foo") == 42


@mark.asyncio
async def test_get_missing(backend: MemoryBackend[int]):
    with raises(KeyError):
        assert await backend.get("foo")


@mark.asyncio
async def test_set_default(backend: MemoryBackend[int]):
    assert await backend.set("foo", 42, if_not_exists=True)
    assert not await backend.set("foo", 43, if_not_exists=True)
    assert await backend.get("foo") == 42
    assert backend.size == 1


@mark.asyncio
async def test_delete_existing(backend: MemoryBackend[int]):
    await backend.set("foo", 42)
    assert await backend.delete("foo")
    with raises(KeyError):
        await backend.get("foo")


@mark.asyncio
async def test_delete_missing(backend: MemoryBackend[int]):
    assert not await backend.delete("foo")


@mark.asyncio
async def test_set_get_many(backend: MemoryBackend[int]):
    await backend.set_many([("foo", 42), ("bar", 100500)])
    assert backend.size == 2
    assert [entry async for entry in backend.get_many("foo", "bar")] == [("foo", 42), ("bar", 100500)]


@mark.asyncio
async def test_set_with_ttl(backend: MemoryBackend[int]):
    with freeze_time("2022-06-11 21:33:00"):
        await backend.set("foo", 42, time_to_live=timedelta(seconds=59))
    with freeze_time("2022-06-11 21:33:58"):
        assert await backend.get("foo") == 42
    with freeze_time("2022-06-11 21:34:00"), raises(KeyError):
        assert await backend.get("foo")
    assert backend.size == 0


@mark.asyncio
async def test_expire_at(backend: MemoryBackend[int]):
    await backend.set("foo", 42)
    await backend.expire_at("foo", datetime(2022, 6, 10, 21, 50, 00, tzinfo=timezone.utc))

    with freeze_time("2022-06-10 21:49:59"):
        assert await backend.get("foo") == 42
    with freeze_time("2022-06-10 21:50:00"), raises(KeyError):
        assert await backend.get("foo")
    assert backend.size == 0


@mark.asyncio
async def test_expire_in(backend: MemoryBackend[int]):
    with freeze_time("2022-06-10 21:49:00"):
        await backend.set("foo", 42)
        await backend.expire_in("foo", timedelta(seconds=59))
    with freeze_time("2022-06-10 21:49:58"):
        assert await backend.get("foo") == 42
    with freeze_time("2022-06-10 21:50:00"), raises(KeyError):
        assert await backend.get("foo")
    assert backend.size == 0


@mark.asyncio
async def test_clear(backend: MemoryBackend[int]):
    await backend.set("foo", 42)
    await backend.clear()
    assert backend.size == 0



from urllib.parse import urlparse

from cachetory.interfaces.backends.sync import SyncBackend

from .dummy import DummyBackend
from .memory import MemoryBackend

try:
    # noinspection PyUnresolvedReferences
    from .redis import RedisBackend
except ImportError:
    _is_redis_available = False
else:
    _is_redis_available = True


def from_url(url: str) -> SyncBackend:
    """
    Creates a synchronous backend from the given URL.

    Examples:
        >>> from_url("redis://localhost:6379")
    """
    parsed_url = urlparse(url)
    scheme = parsed_url.scheme
    if scheme == "memory":
        return MemoryBackend.from_url(url)
    if scheme in ("redis", "rediss", "redis+unix"):
        if not _is_redis_available:
            raise ValueError(f"`{scheme}://` requires `cachetory[redis-sync]` extra")
        return RedisBackend.from_url(url)
    if scheme == "dummy":
        return DummyBackend.from_url(url)
    raise ValueError(f"`{scheme}://` is not supported")



from pytest import mark

from cachetory.backends.async_ import DummyBackend


@mark.asyncio
async def test_get_many():
    async with DummyBackend() as backend:
        assert [item async for item in backend.get_many("foo")] == []



from __future__ import annotations

from typing import Generic

from cachetory.interfaces.serializers import Serializer, ValueT


class NoopSerializer(Serializer[ValueT, ValueT], Generic[ValueT]):
    """
    No-operation serializer: just forwards the value to and from backend.
    """

    @classmethod
    def from_url(cls, url: str) -> NoopSerializer[ValueT]:
        return NoopSerializer[ValueT]()

    def serialize(self, value: ValueT) -> ValueT:
        return value

    def deserialize(self, data: ValueT) -> ValueT:
        return data



from __future__ import annotations

from typing import TypeVar

from typing_extensions import Protocol

from cachetory.interfaces.backends.private import WireT, WireT_co, WireT_contra

ValueT_co = TypeVar("ValueT_co", covariant=True)
"""
Value type returned by a deserializer, should be covariant so that
for A > B we could pass `Deserialize[B]` in place of `Deserialize[A]`.
"""

ValueT_contra = TypeVar("ValueT_contra", contravariant=True)
"""
Value type accepted by a serializer, should be contravariant so that
for A > B we could pass `Serialize[A]` in place of `Serialize[B]`.
"""


class Serialize(Protocol[ValueT_contra, WireT_co]):
    """
    Defines the `serialize` conversion.
    """

    def serialize(self, value: ValueT_contra) -> WireT_co:  # pragma: no cover
        raise NotImplementedError


class Deserialize(Protocol[ValueT_co, WireT_contra]):
    """
    Defines the `deserialize` conversion.
    """

    def deserialize(self, data: WireT_contra) -> ValueT_co:  # pragma: no cover
        raise NotImplementedError


ValueT = TypeVar("ValueT")
"""
Cached value type – this what user «sees» when getting or setting a value in a cache.
"""


class Serializer(Serialize[ValueT, WireT], Deserialize[ValueT, WireT], Protocol[ValueT, WireT]):
    """
    Combines `serialize` and `deserialize` in one protocol.
    """

    @classmethod
    def from_url(cls, url: str) -> Serializer[ValueT, WireT]:  # pragma: no cover
        """
        Instantiate a serializer from the specified URL.

        Returns:
            An instance of specific serializer class.
        """
        raise NotImplementedError



from datetime import datetime, timedelta, timezone
from typing import Optional


def make_deadline(time_to_live: Optional[timedelta] = None) -> Optional[datetime]:
    return datetime.now(timezone.utc) + time_to_live if time_to_live is not None else None



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Dict, Generic, Optional

from cachetory.interfaces.backends.private import WireT
from cachetory.interfaces.backends.sync import SyncBackend
from cachetory.private.datetime import make_deadline


class MemoryBackend(SyncBackend[WireT], Generic[WireT]):
    """
    Memory backend that stores everything in a local dictionary.
    """

    __slots__ = ("_entries",)

    @classmethod
    def from_url(cls, _url: str) -> MemoryBackend:
        return MemoryBackend()

    def __init__(self):
        self._entries: Dict[str, _Entry[WireT]] = {}

    def get(self, key: str) -> WireT:
        return self._get_entry(key).value

    def expire_at(self, key: str, deadline: Optional[datetime]) -> None:
        try:
            entry = self._get_entry(key)
        except KeyError:
            pass
        else:
            entry.deadline = deadline

    def set(
        self,
        key: str,
        value: WireT,
        *,
        time_to_live: Optional[timedelta] = None,
        if_not_exists: bool = False,
    ) -> bool:
        entry = _Entry[WireT](value, make_deadline(time_to_live))
        if if_not_exists:
            return self._entries.setdefault(key, entry) is entry
        else:
            self._entries[key] = _Entry[WireT](value, make_deadline(time_to_live))
            return True

    def delete(self, key: str) -> bool:
        return self._entries.pop(key, _SENTINEL) is not _SENTINEL

    def clear(self) -> None:
        self._entries.clear()

    def _get_entry(self, key: str) -> _Entry[WireT]:
        entry = self._entries[key]
        if entry.deadline is not None and entry.deadline <= datetime.now(timezone.utc):
            self._entries.pop(key, None)  # might get popped by another thread
            raise KeyError(f"`{key}` has expired")
        return entry

    @property
    def size(self) -> int:
        return len(self._entries)


class _Entry(Generic[WireT]):
    """
    `mypy` doesn't support generic named tuples, thus defining this little one.
    """

    value: WireT
    deadline: Optional[datetime]

    __slots__ = ("value", "deadline")

    def __init__(self, value: WireT, deadline: Optional[datetime]):
        self.value = value
        self.deadline = deadline


_SENTINEL = object()



#!/usr/bin/env python

# setuptools doesn't support type hints for now:
# https://github.com/pypa/setuptools/issues/2345
# so we ignoring mypy checks on this package
from setuptools import find_packages, setup  # type: ignore

with open("README.md") as f:
    long_description = f.read()


setup(
    name="cachetory",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Caching library with support for multiple cache backends",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Pavel Perestoronin",
    author_email="pavel.perestoronin@kpn.com",
    url="https://github.com/kpn/cachetory",
    packages=find_packages(exclude=["tests*"]),
    python_requires=">=3.7",
    install_requires=[
        "pydantic>=1.9.0,<2.0.0",
        "typing-extensions>=4.2.0,<5.0.0",
    ],
    extras_require={
        "hiredis-async": ["aioredis[hiredis]>=2.0.1,<3.0.0"],
        "redis-async": ["aioredis>=2.0.1,<3.0.0"],
        "redis-sync": ["redis>=4.3.3,<5.0.0"],
        "zstd": ["zstd>=1.5.2.5,<2.0.0.0"],
    },
    tests_require=["tox"],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
    ],
)



"""
I could define these type variables in each protocol's module,
but it's easier for a reader to come by a central place and find the additional comments.
"""

from typing import TypeVar

WireT_co = TypeVar("WireT_co", covariant=True)
"""
Value type as accepted and returned by backend read operations (as in «data on wire»).
Why not just `bytes`? The generic type would allow implementing e.g. a memory backend
which doesn't necessarily need a serializer and is able to store the values directly.
See, for example, `SyncMemoryBackend` and `NoopSerializer`.
"""

WireT_contra = TypeVar("WireT_contra", contravariant=True)
"""
The contravariance ensures that given A > B: `SyncBackendWrite[A]` < `SyncBackendWrite[B]`,
meaning that user can safely pass `SyncBackendWrite[A]` in place of SyncBackendWrite[B].
And this is needed because `SyncBackendWrite[B]` accepts B as a parameter,
and so should `SyncBackendWrite[A]`.
"""

WireT = TypeVar("WireT")
"""
Invariant backend value type – exact type which is stored at backend side.
It is independent from a cached value type. Backend value is usually
a serialized cached value.
"""



from __future__ import annotations

import pickle
from typing import Generic
from urllib.parse import parse_qsl, urlparse

from pydantic import BaseModel, Field, conint

from cachetory.interfaces.serializers import Serializer, ValueT


class PickleSerializer(Serializer[ValueT, bytes], Generic[ValueT]):
    """
    Uses the standard built-in `pickle` serialization.
    """

    __slots__ = ("protocol",)

    @classmethod
    def from_url(cls, url: str) -> PickleSerializer[ValueT]:
        params = _UrlParams.parse_obj(dict(parse_qsl(urlparse(url).query)))
        return cls(pickle_protocol=params.pickle_protocol)

    def __init__(self, pickle_protocol: int = pickle.HIGHEST_PROTOCOL):
        self.protocol = pickle_protocol

    def serialize(self, value: ValueT) -> bytes:
        return pickle.dumps(value, self.protocol)

    def deserialize(self, data: bytes) -> ValueT:
        return pickle.loads(data)


class _UrlParams(BaseModel):
    pickle_protocol: conint(ge=0, le=pickle.HIGHEST_PROTOCOL) = Field(pickle.HIGHEST_PROTOCOL, alias="pickle-protocol")  # type: ignore



from __future__ import annotations

from typing import Any, Generic, Iterable, cast
from urllib.parse import urlparse

from cachetory.interfaces.backends.private import WireT
from cachetory.interfaces.serializers import Serializer, ValueT
from cachetory.serializers.compressors.zlib import ZlibCompressor
from cachetory.serializers.noop import NoopSerializer
from cachetory.serializers.pickle import PickleSerializer

try:
    # noinspection PyUnresolvedReferences
    from cachetory.serializers.compressors import ZstdCompressor
except ImportError:
    _is_zstd_available = False
else:
    _is_zstd_available = True


class ChainedSerializer(Serializer[ValueT, WireT], Generic[ValueT, WireT]):
    """
    Sequentially applies the chain of serializers.
    Allows defining multiple steps of serialization.
    """

    __slots__ = ("_layers",)

    @classmethod
    def from_url(cls, url: str) -> ChainedSerializer[ValueT, WireT]:
        parsed_url = urlparse(url)
        schemes = parsed_url.scheme.split("+")
        return cls(cls._make_layer(scheme, url) for scheme in schemes)

    def __init__(self, layers: Iterable[Serializer]):
        self._layers = list(layers)

    def serialize(self, value: ValueT) -> WireT:
        value = cast(Any, value)
        for layer in self._layers:
            value = layer.serialize(value)
        return cast(WireT, value)

    def deserialize(self, data: WireT) -> ValueT:
        value = cast(Any, data)
        for layer in self._layers[::-1]:
            value = layer.deserialize(value)
        return cast(ValueT, value)

    @classmethod
    def _make_layer(cls, scheme: str, url: str) -> Serializer:
        if scheme == "pickle":
            return PickleSerializer.from_url(url)
        if scheme in ("noop", "null"):
            return NoopSerializer.from_url(url)
        if scheme in ("zstd", "zstandard"):
            if not _is_zstd_available:
                raise ValueError(f"`{scheme}://` requires `cachetory[zstd]` extra")
            return ZstdCompressor.from_url(url)
        if scheme == "zlib":
            return ZlibCompressor.from_url(url)
        raise ValueError(f"`{scheme}://` is not supported")
